import click

import boo.downloader as downloader


@click.group()
def cli():
    """Download 2012-2018 corporate annual financial statements from Rosstat."""
    pass


@cli.command()
@click.argument("year")
def url(year: str):
    """Show URL for remote zipfile."""
    year = accept_year(year)
    click.echo(downloader.RemoteZipFile(year).url, nl=False)


def accept_year(year):
    year = int(year)
    if year in downloader.available_years():
        return year
    else:
        msg = f"{year} is not a valid YEAR, must be in 2012-2018 range."
        click.echo(msg, err=True)
        return None


def on_existing_file(year: int, remote_zip, local_zip):
    click.echo(f"{year} file already exists: {local_zip}")
    if remote_zip.is_same_size(local_zip):
        click.echo("Remote file size equals local file size")
    else:
        click.echo("Found local file of wrong size, delete recommended")


def to_terminal(*args):
    xs = " ".join([str(x) for x in args])
    return click.echo(xs)


@cli.command()
@click.argument("year")
def download(year):
    """Download zip file to local computer."""
    year = accept_year(year)
    if year:
        downloader.download(year, folder=None, echo=to_terminal)


@cli.command()
@click.argument("year")
def unpack(year):
    """Unpack local zip file to get csv file."""
    year = accept_year(year)
    if year:
        downloader.unpack(year, folder=None, echo=to_terminal)


@cli.command()
def years():
    """List available years."""
    click.echo(" ".join([str(x) for x in downloader.available_years()]))


@cli.command()
@click.argument("year", nargs=-1)
def files(year):
    """Show default paths to local zip and csv file by year."""
    click.echo(downloader.default_data_folder())
    if year:
        year = accept_year(year[0])  # check again nargs=-1 makes input a tuple
        if year:
            click.echo(downloader.path_zip(year))
            click.echo(downloader.path_csv(year))


if __name__ == "__main__":
    cli()



import pandas as pd

from boo.columns import INDEX, NAMES
from boo.dataframe import canonic_df
from boo.downloader import path_csv


def read_intermediate_df(year, directory=None, **user_args):
    """Read intermeiate data for *year* as dataframe.

    You can use read_intermediate_df(year, nrows=100) for small example.

    Returns:
        pandas.DataFrame
    """
    kwargs = dict(
        encoding="windows-1251",
        sep=";",
        header=None,
        usecols=INDEX,  # read only a subset of columns
        names=[x for x in NAMES.keys()],  # give new names to these columns
        dtype=NAMES,  # enforce string or int types
    )
    return pd.read_csv(path_csv(year, directory), **kwargs, **user_args)


def read_dataframe(year: int, directory=None, **kwargs):
    """Read canonic data for *year* as dataframe.

    Returns:
        pandas.DataFrame
    """
    return canonic_df(read_intermediate_df(year, directory, **kwargs))



"""Availabale years and URLs for them."""

from dataclasses import dataclass
from pathlib import Path
from typing import List, Optional, Union
from urllib.request import urlopen
from zipfile import ZipFile

import requests
from tqdm import tqdm

# Note: must manually hardcode new timestamps when new version of
# data arrives.


def timestamps(year: int):
    return {
        2012: ("7708234640-", "20200331"),  # different
        2013: ("7708234640-", "20200331"),  # different
        2014: ("7708234640-", "20200327"),  # similar
        2015: ("7708234640-", "20200327"),  # similar
        2016: ("7708234640-", "20200327"),  # similar
        2017: ("7708234640-", "20200327"),  # similar
        2018: ("7708234640-7708234640", "20200327"),  # different
    }[year]


def required_file_size(year):
    return {
        0: 4_501,
        2012: 88_169_551,
        2013: 195_986_142,
        2014: 225_816_835,
        2015: 262_173_543,
        2016: 268_515_446,
        2017: 273_328_344,
        2018: 269_452_075,
    }[year]


def size(path):
    return path.stat().st_size


def to_megabytes(b: int, precision: int = 1):
    return round(b / 2 ** (10 * 2), precision)


def is_valid_size(path, year):
    return size(path) == required_file_size(year)


def filename(year):
    if year == 0:
        return "sample"
    _, timestamp = timestamps(year)
    return f"data-{timestamp}-structure-{year}1231"


def csv_filename(year):
    return filename(year) + ".csv"


def zip_filename(year):
    return filename(year) + ".zip"


def mk_url_pure(year):
    mid, _ = timestamps(year)
    return f"https://rosstat.gov.ru/opendata/{mid}bdboo{year}/" + zip_filename(year)


def mk_url(year):
    if year == 0:  # for testing
        return (
            "https://github.com/ru-corporate/boo/blob/master/assets/sample.zip?raw=true"
        )
    return mk_url_pure(year)


def available_years() -> List[int]:
    """List available years with datasets."""
    return [2012, 2013, 2014, 2015, 2016, 2017, 2018]


def curl(path: str, url: str, n_chunks=None):
    r = requests.get(url, stream=True)
    iterable = tqdm(r.iter_content(chunk_size=1024), unit=" k")
    with open(path, "wb") as f:
        i = 0
        for chunk in iterable:
            if chunk:  # filter out keep-alive new chunks
                f.write(chunk)
            i += 1
            if n_chunks and i >= n_chunks:
                break


def url_content_length(url: str) -> int:
    """Get size of a remote file at url."""
    with urlopen(url) as f:
        return int(f.info()["Content-Length"])


def file_length(year: int):
    """Get size of a remote url by year."""
    return url_content_length(url=mk_url(year))


def default_data_folder() -> Path:
    home = Path.home() / ".boo"
    home.mkdir(exist_ok=True)
    return home


def unzip(path, folder=None):
    folder = folder if folder else default_data_folder()
    with ZipFile(path) as zf:
        zf.extractall(folder)
        return zf.namelist()


def persist(folder: Union[Path, str, None]) -> Path:
    return Path(folder) if folder else default_data_folder()


def path_zip(year: int, folder: Optional[Path] = None) -> Path:
    return persist(folder) / f"{year}.zip"


def path_csv(year: int, folder: Optional[Path] = None) -> Path:
    return persist(folder) / csv_filename(year)


@dataclass
class RemoteZipFile:
    year: int

    def __post_init__(self):
        self._size = None

    @property
    def url(self):
        return mk_url(self.year)

    def size(self):
        if self._size is None:
            self._size = url_content_length(self.url)
        return self._size

    def download(self, destination_folder):
        return curl(path_zip(self.year, destination_folder), self.url)

    def validate_size(self):
        a = self.size()
        b = required_file_size(self.year)
        if a != b:
            raise ValueError(f"Remote file size {a} does not match {b}")

    def is_same_size(self, local_file: Path):
        return size(local_file) == self.size()


def download(year, directory: Optional[str] = None, echo=print):
    folder = persist(directory)
    remote_zip = RemoteZipFile(year)
    remote_zip.validate_size()
    echo("Year", year, "remote file size is", remote_zip.size())
    local_zip = path_zip(year, folder)
    if local_zip.exists() and remote_zip.is_same_size(local_zip):
        echo("File already downloaded:", local_zip)
    else:
        echo("Downloading for year", year)
        remote_zip.download(folder)
        echo("Downloaded zip file", local_zip)
        echo("Size", size(local_zip))
    return str(local_zip)


def unpack(year, directory: Optional[str] = None, echo=print):
    folder = persist(directory)
    local_csv = path_csv(year, folder)
    if local_csv.exists():
        echo("File", local_csv, "already exists")
    else:
        local_zip = path_zip(year, folder)
        echo("Unpacking local zip file for year", year, "to folder", folder)
        files = unzip(local_zip, folder)
        if files[0] == csv_filename(year):
            local_csv = path_csv(year, folder)
            echo("Unpacked", local_csv)
            echo("Size", local_csv)
        else:
            echo("Unpacked several files:", " ".join(files))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['boo', 'boo.dataframe', 'boo.okved']

package_data = \
{'': ['*']}

install_requires = \
['click>=7.1.2,<8.0.0',
 'pandas==1.1.5',
 'requests>=2.25.1,<3.0.0',
 'tqdm>=4.59.0,<5.0.0']

setup_kwargs = {
    'name': 'boo',
    'version': '0.2.0',
    'description': 'Russian corporate financial reports 2012-2018.',
    'long_description': None,
    'author': 'Evgeniy Pogrebnyak',
    'author_email': 'e.pogrebnyak@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
